

#if	defined(HASNCACHE)
/*
 * rnch.frag - read Sun format (struct ncache) name cache
 *
 * This fragment is effective only when HASNCACHE is defined.
 */

/*
 * The caller must:
 *
 *	#include the relevant header file -- e.g., <sys/dnlc.h>.
 *
 *	Define X_NCSIZE as the Nl[] index to the kernel cache size
 *	variable, or, if X_NCSIZE is undefined, define FIXED_NCSIZE
 *	as the size of the kernel cache.
 *
 *	Define X_NCACHE as the Nl[] index to the kernel cache address.
 *	and define ADDR_NCACHE if Nl[X_NCACHE].n_value is the address
 *	of the cache, rather than the address of a pointer to it.
 *
 *	Define NCACHE_NXT if the kernel's name cache is a linked
 *	list starting at Nl[X_NCACHE].n_value, rather than a table
 *	starting at that address.
 *
 * The caller may:
 *
 *	Define NCACHE_NODEID	as the name of the element in the
 *				ncache structure that contains the
 *				vnode's capability ID.
 *
 *	Define NCACHE_PARID	as the name of the element in the
 *				ncache structure that contains the
 *				parent vnode's capability ID.
 *
 * Note: if NCACHE_NODEID is defined, then NCACHE_PARID must be defined.
 *
 *
 * The caller may need to:
 *
 *	Define this prototype for ncache_load() if it is called
 *	before rnch.frag is included.
 *
 *		_PROTOTYPE(static void ncache_load,(void));
 */


/*
 * Local static values
 */

static int Mch;				/* name cache hash mask */
static int Nc = 0;			/* size of name cache */
static int Nch = 0;			/* size of name cache hash pointer
					 * table */
struct l_nch {
	struct vnode *vp;		/* vnode address */
	union {
		struct vnode *dp;	/* parent vnode address */
		struct l_nch *pa;	/* parent Ncache address */
	} u;

# if	defined(NCACHE_NODEID)
	unsigned long id;		/* node's capability ID */
	unsigned long did;		/* parent node's capability ID */
# endif	/* defined(NCACHE_NODEID) */

	char nm[NC_NAMLEN+1];		/* name */
	int nl;				/* name length */
};

static struct l_nch *Ncache = NULL;	/* the local name cache */
static struct l_nch **Nchash = NULL;	/* Ncache hash pointers */

# if	defined(NCACHE_NODEID)
_PROTOTYPE(static struct l_nch *ncache_addr,(unsigned long i, struct vnode *v));
#define ncachehash(i,v)		Nchash+(((((int)(v)>>2)+((int)(i)))*31415)&Mch)
# else	/* !defined(NCACHE_NODEID) */
_PROTOTYPE(static struct l_nch *ncache_addr,(struct vnode *v));
#define ncachehash(v)		Nchash+((((int)(v)>>2)*31415)&Mch)
# endif	/* defined(NCACHE_NODEID) */


/*
 * ncache_addr() - look up a node's local ncache address
 */

static struct l_nch *

# if	defined(NCACHE_NODEID)
ncache_addr(i, v)
# else	/* !defined(NCACHE_NODEID) */
ncache_addr(v)
# endif	/* defined(NCACHE_NODEID) */

# if	defined(NCACHE_NODEID)
	unsigned long i;			/* capability ID */
# endif	/* defined(NCACHE_NODEID) */

	struct vnode *v;			/* vnode's address */
{
	struct l_nch **hp;

# if	defined(NCACHE_NODEID)
	for (hp = ncachehash(i, v); *hp; hp++)
# else	/* !defined(NCACHE_NODEID) */
	for (hp = ncachehash(v); *hp; hp++)
# endif	/* defined(NCACHE_NODEID) */

	{

# if	defined(NCACHE_NODEID)
		if ((*hp)->vp == v && (*hp)->id == i)
# else	/* !defined(NCACHE_NODEID) */
		if ((*hp)->vp == v)
# endif	/* defined(NCACHE_NODEID) */

			return(*hp);
	}
	return(NULL);
}


/*
 * ncache_load() - load the kernel's name cache
 */

static void
ncache_load()
{
	int i, len, n;
	struct l_nch **hp, *lc;
	struct ncache *kc;
	KA_T kp;

# if	defined(NCACHE_NXT)
	struct ncache nc;
# else	/* !defined(NCACHE_NXT) */
	struct ncache *kca;
# endif	/* defined(NCACHE_NXT) */

	if (!Fncache)
		return;
/*
 * Establish kernel cache size.
 */

# if	defined(X_NCSIZE)
	if (!Nl[X_NCSIZE].n_value
	||  kread((KA_T)Nl[X_NCSIZE].n_value, (char *)&Nc, sizeof(Nc)))
	{
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: WARNING: can't read name cache size: %#x\n",
		    Pn, Nl[X_NCSIZE].n_value);
	    Nc = 0;
	    return;
	}
# else	/* !defined(X_NCSIZE) */

	Nc = FIXED_NCSIZE;

# endif	/* defined(X_NCSIZE) */

	if (Nc < 1) {
	
no_cache:

	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: WARNING: unusable name cache size: %d\n", Pn, Nc);
	    Nc = 0;
	    return;
	}
/*
 * Establish kernel cache address.
 */

# if	defined(ADDR_NCACHE)
	if (!(kp = Nl[X_NCACHE].n_value)) {
	    if (!Fwarn)
		(void) fprintf(stderr, "%s: WARNING: no name cache address\n",
			Pn);
	    Nc = 0;
	    return;
	}
# else	/* !defined(ADDR_NCACHE) */
	if (!Nl[X_NCACHE].n_value
	||  kread((KA_T)Nl[X_NCACHE].n_value, (char *)&kp, sizeof(kp))) {
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: WARNING: can't read name cache pointer: %#x\n",
		    Pn, Nl[X_NCACHE].n_value);
	    Nc = 0;
	    return;
	}
# endif	/* defined(ADDR_NCACHE) */

/*
 * Allocate space for a local copy of the kernel's cache and
 * fill it from kernel memory.
 */

# if	!defined(NCACHE_NXT)
	len = Nc * sizeof(struct ncache);
	if ((kca = (struct ncache *)malloc((MALLOC_S)len)) == NULL) {
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: can't allocate name cache space: %d\n", Pn, len);
	    exit(1);
	}
	if (kread(kp, (char *)kca, len)) {
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: WARNING: can't read kernel's name cache: %#x\n",
		    Pn, kp);
	    Nc = 0;
	    return;
	}
# endif	/* !defined(NCACHE_NXT) */

/*
 * Allocate space for a local version of the cache.
 */
	len = Nc * sizeof(struct l_nch);
	if ((Ncache = (struct l_nch *)malloc(len)) == NULL) {

no_local_space:

	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: no space for %d byte local name cache\n", Pn, len);
	    exit(1);
	}
/*
 * Build a local copy of the kernel name cache.
 */

# if	defined(NCACHE_NXT)
	for (i = n = 0, kc = &nc, lc = Ncache; i < Nc && kp; i++)
# else	/* !defined(NCACHE_NXT) */
	for (i = n = 0, kc = kca, lc = Ncache; i < Nc; i++, kc++)
# endif	/* defined(NCACHE_NXT) */

	{

# if	defined(NCACHE_NXT)
		if (kread(kp, (char *)kc, sizeof(nc)))
			break;
		kp = (KA_T)kc->NCACHE_NXT;
# endif	/* defined(NCACHE_NXT) */

		if (!kc->vp)
			continue;
		if ((len = kc->namlen) < 1 || len > NC_NAMLEN)
			continue;
		if (len < 3 && kc->name[0] == '.') {
			if (len == 1 || (len == 2 && kc->name[1] == '.'))
				continue;
		}
		lc->vp = kc->vp;
		lc->u.dp = kc->dp;
		(void) strncpy(lc->nm, kc->name, len);
		lc->nm[len] = '\0';
		lc->nl = strlen(lc->nm);

# if	defined(NCACHE_NODEID)
		lc->id = (unsigned long)kc->NCACHE_NODEID;
		lc->did = (unsigned long)kc->NCACHE_PARID;
# endif	/* defined(NCACHE_NODEID) */

		n++;
		lc++;
	}
/*
 * Reduce the local name cache memory allocation, if possible.
 * Free the space allocated to the kernel name cache copy.
 */

# if	!defined(NCACHE_NXT)
	(void) free((FREE_P *)kca);
# endif	/* !defined(NCACHE_NXT) */

	if (n < 1) {
		Nc = n;
		(void) free((FREE_P *)Ncache);
		Ncache = NULL;
		goto no_cache;
	}
	if (n < Nc) {
		Nc = n;
		len = Nc * sizeof(struct l_nch);
		if ((Ncache = (struct l_nch *)realloc(Ncache, len)) == NULL)
			goto no_local_space;
	}
/*
 * Build a hash table to locate Ncache entries.
 */
	for (Nch = 1; Nch < Nc; Nch <<= 1)
		;
	Nch <<= 1;
	Mch = Nch - 1;
	if ((Nchash = (struct l_nch **)calloc(Nch + Nc, sizeof(struct l_nch *)))
	== NULL) {
	    if (!Fwarn)
		(void) fprintf(stderr,
		    "%s: no space for %d name cache hash pointers\n",
		    Pn, Nch + Nc);
	    exit(1);
	}
	for (i = 0, lc = Ncache; i < Nc; i++, lc++) {

# if	defined(NCACHE_NODEID)
		for (hp = ncachehash(lc->id, lc->vp), n = 1; *hp; hp++)
# else	/* !defined(NCACHE_NODEID) */
		for (hp = ncachehash(lc->vp), n = 1; *hp; hp++)
# endif	/* defined(NCACHE_NODEID) */

		{
		    if ((*hp)->vp == lc->vp
		    &&  strcmp((*hp)->nm, lc->nm) == 0
		    &&  (*hp)->u.dp == lc->u.dp

# if	defined(NCACHE_NODEID)
		    &&  (*hp)->id == lc->id && (*hp)->did == lc->did
# endif	/* defined(NCACHE_NODEID) */

		    ) {
			n = 0;
			break;
		    }
		}
		if (n)
		    *hp = lc;
	}
/*
 * Make a final pass through the local cache and convert parent vnode
 * addresses to local name cache pointers.
 */
	for (i = 0, lc = Ncache; i < Nc; i++, lc++) {
		if (!lc->u.dp)
			continue;

# if	defined(NCACHE_NODEID)
		lc->u.pa = ncache_addr(lc->did, lc->u.dp);
# else	/* !defined(NCACHE_NODEID) */
		lc->u.pa = ncache_addr(lc->u.dp);
# endif	/* defined(NCACHE_NODEID) */
	
	}
}


/*
 * ncache_lookup() - look up a node's name in the kernel's name cache
 */

char *
ncache_lookup(buf, blen)
	char *buf;			/* receiving name buffer */
	int blen;			/* receiving buffer length */
{
	char *cp;
	struct l_nch *lc;
	int nl, rlen;
/*
 * Look up the name cache entry for the node address.
 */
	if (Nc == 0

# if	defined(NCACHE_NODEID)
	||  (lc = ncache_addr(Lf->id, (struct vnode *)Lf->na)) == NULL)
# else	/* !defined(NCACHE_NODEID) */
	||  (lc = ncache_addr((struct vnode *)Lf->na)) == NULL)
# endif	/* defined(NCACHE_NODEID) */

		return(NULL);
	if ((nl = lc->nl) > (blen - 1))
		return(NULL);
	cp = buf + blen - nl - 1;
	rlen = blen - nl - 1;
	(void) strcpy(cp, lc->nm);
/*
 * Look up the name cache entries that are parents of the node address.
 * Quit when:
 *
 *	there's no parent;
 *	the name is too large to fit in the receiving buffer.
 */
	for (;;) {
		if (!(lc = lc->u.pa) || ((nl = lc->nl) + 1) > rlen)
			break;
		*(cp - 1) = '/';
		cp--;
		rlen--;
		(void) strncpy((cp - nl), lc->nm, nl);
		cp -= nl;
		rlen -= nl;
	}
	return(cp);
}
#endif	/* defined(HASNCACHE) */
