/*
 * dnode.c - SCO Unix node functions for lsof
 */


/*
 * Copyright 1995 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1995 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dnode.c,v 1.5 95/03/05 07:56:58 abe Exp $";
#endif


#include "lsof.h"


_PROTOTYPE(static struct dev * lkupdev,(dev_t *dev, int stream));


/*
 * lkupdev() - look up device by device number
*/

static struct dev *
lkupdev(dev, stream)
	dev_t *dev;			/* device */
	int stream;			/* stream if 1 */
{
	struct clone *c;
	int low, hi, mid;
	struct dev *dp;

	if (!Sdev)
		readdev();
/*
 * Search for clone.
 */
	if (stream && Clone) {
		for (c = Clone; c; c = c->next) {
			if (major(*dev) == minor(Devtp[c->dx].rdev))
				return(&Devtp[c->dx]);
		}
	}
/*
 * Search device table for match.
 */
	low = mid = 0;
	hi = Ndev - 1;
        while (low <= hi) {
                mid = (low + hi) / 2;
		dp = Sdev[mid];
		if (*dev == dp->rdev)
			return(dp);
		else if (*dev < dp->rdev)
			hi = mid - 1;
		else
                        low = mid + 1;
        }
	return(NULL);
}


/*
 * process_node() - process node
 */

void
process_node(na)
	caddr_t na;			/* inode kernel space address */
{
	char *cp, dev_ch[32];
	short dl;
	struct filock fl;
	struct inode i;
	short ity, j, k, l;
	KA_T ka, qp;
	struct mounts *lm;
	struct module_info mi;
	int prdev;
	struct rnode r;
	struct qinit qi;
	struct stdata sd;
	struct queue q;
	char *tn;
	int type;
	struct dev *dp;
/*
 * Read the inode.
 */
	if ( ! na) {
		enter_nm("no inode address");
		return;
	}
	if (readinode((struct inode *)na, &i)) {
		enter_nm(Namech);
		return;
	}
/*
 * Identify the node type.
 */
	ity = i.i_fstyp;
	type = i.i_ftype & IFMT;
	if (ity < 1 || ity > Fsinfomax || Fsinfo[ity-1] == NULL) {
		(void) sprintf(Namech, "unknown fstyp (%d) in inode", ity);
		enter_nm(Namech);
		return;
	}
	if (strcasecmp(Fsinfo[ity-1], "HS") == 0)
		Ntype = N_HSFS;
	 else if (strcasecmp(Fsinfo[ity-1], "NFS") == 0) {

	/*
	 * Get information on NFS file.
	 */
		Ntype = N_NFS;
		Lf->is_nfs = 1;
		if (Fnfs)
			Lf->sf |= SELNFS;
		if (i.i_fsptr == NULL || readrnode((caddr_t)i.i_fsptr, &r)) {
			(void) sprintf(Namech, "can't read rnode (%#x)",
			    i.i_fsptr);
			enter_nm(Namech);
			return;
		}
	} else if (HaveSockdev && (i.i_ftype & IFMT) == IFCHR
	       &&  major(i.i_rdev) == Sockdev)
	{

	/*
	 * Process a socket.
	 */
		process_socket(&i);
		return;
	} else {

	/*
	 * Determine the node type from the inode file type.
	 */
		switch (type) {

		case IFBLK:
			Ntype = N_BLK;
			break;
		case IFCHR:
			Ntype = N_CHR;
			break;
		case IFIFO:
			Ntype = N_FIFO;
			break;
		case IFMPB:
		case IFMPC:
			Ntype = N_MPC;
			break;
		}
	}
/*
 * Obtain lock information.
 */
	if (i.i_filocks) {
		if ((kread((KA_T)i.i_filocks, (char *)&fl, sizeof(fl)))) {
			(void) sprintf(Namech, "can't read filock (%#x)",
				i.i_filocks);
			enter_nm(Namech);
			return;
		} else {
			if (fl.set.l_type == F_RDLCK)
				Lf->lock = 'R';
			else if (fl.set.l_type == F_WRLCK)
				Lf->lock = 'W';
			else if (fl.set.l_type == (F_RDLCK | F_WRLCK))
				Lf->lock = '*';
		}
	}
/*
 * Determine the device.
 */
	switch (Ntype) {
	case N_BLK:
	case N_FIFO:
	case N_HSFS:
	case N_REGLR:
		Lf->dev = i.i_dev;
		Lf->dev_def = 1;
		break;
	case N_CHR:
		Lf->dev = i.i_rdev;
		Lf->dev_def = 1;
		if (i.i_sptr) {

		/*
		 * This is a stream, but since the device name is placed
		 * in Namech here, Lf->is_stream is not set, thus preventing
		 * printname() from trying to do it, too.  Namech is:
		 *    /dev/* name if it exists for i.i_rdev;
		 *    cdevsw[].d_name if it exists for major(i.i_rdev);
		 *    "stream" otherwise.
		 */
		    (void) strcpy(Namech, "STR:");
		    k = strlen(Namech);
		    cp = NULL;
		    if ((dp = lkupdev(&i.i_rdev, 1)) != NULL) {
			(void) strcpy(&Namech[k], dp->name);
			k += strlen(dp->name);
			if ((cp = strrchr(dp->name, '/')) != NULL)
			    cp++;
		    } else if ((j = major(i.i_rdev)) < Cdevcnt
			   &&  (cp = Cdevsw[j]) != NULL)
		    {
			(void) strcpy(Namech, cp);
			k += strlen(cp);
		    }
		/*
		 * Get the module names of all queue elements of the stream's
		 * sd_wrq queue.  Skip module names that end in "head", 
		 * match the last component of the /dev name, or match the
		 * cdevsw[].d_name.
		 */
		    if (kread((KA_T)i.i_sptr, (char *)&sd, sizeof(sd)) == 0) {
			dl = sizeof(dev_ch) - 1;
			dev_ch[dl] = '\0';
			qp = (KA_T)sd.sd_wrq;
			for (j = 0; qp && j < 20; j++, qp = (KA_T)q.q_next) {
			    if (kread(qp, (char *)&q, sizeof(q)))
				break;
			    if ((ka = (KA_T)q.q_qinfo) == (KA_T)NULL
			    ||  kread(ka, (char *)&qi, sizeof(qi)))
				continue;
			    if ((ka = (KA_T)qi.qi_minfo) == (KA_T)NULL
			    ||  kread(ka, (char *)&mi, sizeof(mi)))
				continue;
			    if ((ka = (KA_T)mi.mi_idname) == (KA_T)NULL
			    ||  kread(ka, dev_ch, dl))
				continue;
			    if ((l = strlen(dev_ch)) < 1)
				continue;
			    if (l >= 4 && strcmp(&dev_ch[l - 4], "head") == 0)
				continue;
			    if (cp && strcmp(cp, dev_ch) == 0)
				continue;
			    if (k) {
				if ((k + 2) > (MAXPATHLEN - 1))
				    break;
				(void) strcpy(&Namech[k], "->");
				k += 2;
			    }
			    if ((k + l) > (MAXPATHLEN - 1))
				break;
			    (void) strcpy(&Namech[k], dev_ch);
			    k += l;
			}
		    }
		} else {
			if (strcasecmp(Fsinfo[ity-1], "COM") == 0) {
				Ntype = N_COM;
				Lf->is_com = 1;
			} else
				Lf->is_chr_dev  = 1;
		}
		break;
	case N_NFS:
		Lf->dev = (dev_t)_makedev(~major(i.i_dev), minor(i.i_dev));
		Lf->dev_def = 1;
		break;
	}
/*
 * Determine the inode number.
 */
	switch (Ntype) {
	case N_HSFS:

	/*
	 * For some reason the High Sierra inode number, as reported by
	 * "ls -i" and stat(2), is the lower 16 bits of i_number.
	 */
		Lf->inode = (unsigned long)(i.i_number & 0xffff);
		Lf->inp_ty = 1;
		break;
	case N_NFS:
		Lf->inode = (unsigned long)r.r_fh.fh_fno;
		Lf->inp_ty = 1;
		break;
	case N_BLK:
	case N_CHR:
	case N_COM:
	case N_FIFO:
	case N_REGLR:
		Lf->inode = (unsigned long)i.i_number;
		Lf->inp_ty = 1;
		break;
	}
/*
 * Determine the file size.
 */
	if (Foffset)
		Lf->off_def = 1;
	else {
		switch (Ntype) {
		case N_BLK:
			Lf->sz = i.i_size;
			Lf->sz_def = 1;
			break;
		case N_CHR:
		case N_COM:
			if (!Fsize)
				Lf->off_def = 1;
			break;
		case N_FIFO:
			if (!Fsize)
				Lf->off_def = 1;
			break;
		case N_HSFS:
		case N_NFS:
			Lf->sz = i.i_size;
			Lf->sz_def = 1;
			break;
		case N_REGLR:
			if (type == IFREG || type == IFDIR)
				Lf->sz = i.i_size;
				Lf->sz_def = 1;
			break;
		}
	}
/*
 * Format the type name.
 */
	switch (type) {
	case IFDIR:
		tn = "DIR";
		break;
	case IFBLK:
		tn = "BLK";
		break;
	case IFCHR:
		tn = "CHR";
		break;
	case IFREG:
		tn = "REG";
		break;
	case IFMPC:
		tn = "MPC";
		break;
	case IFMPB:
		tn = "MPB";
		break;
	case IFIFO:
		tn = "FIFO";
		break;

#if	defined(IFLNK)
	case IFLNK:
		tn = "LINK";
		break;
#endif	/* defined(IFLNK) */

	default:
		if (type > 9999)
			(void) sprintf(Lf->type, "*%03d", type % 1000);
		else
			(void) sprintf(Lf->type, "%4d", type);
		(void) strcpy(Namech, "unknown inode type");
		tn = NULL;
	}
	if (tn)
		(void) strcpy(Lf->type, tn);
/*
 * Save the file system names.
 */
	switch (Ntype) {
	case N_BLK:
	case N_FIFO:
	case N_HSFS:
	case N_NFS:
	case N_REGLR:
		if (Lf->dev_def) {
			for (lm = Mtab; lm; lm = lm->next) {
				if (Lf->dev == lm->dev) {
					Lf->fsdir = lm->dir;
					Lf->fsdev = lm->fsname;
					break;
				}
			}
		}
		break;
	}
/*
 * If this is a IFCHR file and it's missing an inode number, try to
 * supply one.
 */
	if (Lf->inp_ty == 0 && type == IFCHR && Lf->dev_def)
		find_ch_ino();
/*
 * Test for specified file.
 */
	if (Sfile && is_file_named(NULL, type))
		Lf->sf |= SELNM;
/*
 * Enter name characters.
 */
	if (Namech[0])
		enter_nm(Namech);
}
