/*
 * dsock.c - SCO Unix socket processing functions for lsof
 */


/*
 * Copyright 1995 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1995 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dsock.c,v 1.2 95/01/21 10:20:38 abe Exp $";
#endif


#include "lsof.h"


/*
 * process_socket() - process socket
 */

void
process_socket(i)
	struct inode *i;		/* inode pointer */
{
	struct domain d;
	char dev_ch[32];
	struct inpcb pcb;
	short pcbs, shs;
	struct socket s;
	KA_T sa, spa;
	struct stdata sd;
	struct queue sh;
	struct sockaddr_in *si;
	
	(void) strcpy(Lf->type, "sock");
	pcbs = shs = 0;
/*
 * Read socket.
 */
	if (Socktab == (KA_T)NULL) {
		(void) enter_nm("No kernel socket table");
		return;
	}
	spa = Socktab + (minor(i->i_rdev) * sizeof(struct socket *));
	if (kread(spa, (char *)&sa, sizeof(sa))) {
		(void) sprintf(Namech, "can't read socket pointer at %#x",
			spa);
		enter_nm(Namech);
	}
	if (kread(sa, (char *)&s, sizeof(s))) {
		(void) sprintf(Namech, "can't read socket structure at %#x",
			sa);
		enter_nm(Namech);
		return;
	}
/*
 * Read domain structure.
 */
	if (s.so_proto.pr_domain == (struct domain *)NULL
	||  kread((KA_T)s.so_proto.pr_domain, (char *)&d, sizeof(d))) {
	    (void) sprintf(Namech, "can't read protocol domain from %#x",
		s.so_proto.pr_domain);
	    enter_nm(Namech);
	    return;
	}
/*
 * Process by protocol domain.
 */
	switch (d.dom_family) {
	case AF_INET:
	    if (Fnet)
		Lf->sf |= SELNET;
	    (void) strcpy(Lf->type, "inet");
	    printiproto((int)s.so_proto.pr_protocol);
	    Lf->inp_ty = 2;
	/*
	 * Get protocol control block address from stream head queue structure.
	 */
	    if (s.so_stp
	    &&  readstdata((struct stdata *)s.so_stp, &sd) == 0
	    &&  readsthead(sd.sd_wrq, &sh) == 0)
		shs = 1;
	    if (shs && sh.q_ptr) {
		(void) sprintf(dev_ch, "0x%08x", sh.q_ptr);
		enter_dev_ch(dev_ch);
		if (kread((KA_T)sh.q_ptr, (char *)&pcb, sizeof(pcb)) == 0)
		    pcbs = 1;
	    }
	/*
	 * Print local and remote addresses.
	 */
	    if (pcbs) {
		printinaddr(&pcb.inp_laddr, (int)ntohs(pcb.inp_lport));
		if (pcb.inp_faddr.s_addr != INADDR_ANY || pcb.inp_fport != 0) {
		    (void) strcat(endnm(), "->");
		    printinaddr(&pcb.inp_faddr, (int)ntohs(pcb.inp_fport));
		}
	    } else {
		si = (struct sockaddr_in *)&s.so_name;
		printinaddr(&si->sin_addr, (int)ntohs(si->sin_port));
		si = (struct sockaddr_in *)&s.so_peer;
		if (si->sin_addr.s_addr != INADDR_ANY || si->sin_port != 0) {
		    (void) strcat(endnm(), "->");
		    printinaddr(&si->sin_addr, (int)ntohs(si->sin_port));
		}
	    }
	/*
	 * Set size or offset.
	 */
	    if (shs) {
		Lf->sz = sh.q_count;
		Lf->sz_def = 1;
	    } else
		Lf->off_def = 1;
	    break;
	default:
	    (void) sprintf(Namech, "unknown protocol %d",
		s.so_proto.pr_protocol);
	}
	enter_nm(Namech);
}
