/*
 * dnode.c - SunOS (Solaris 1.x and 2.x) node reading functions for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: dnode.c,v 1.20 95/02/13 14:48:32 abe Exp $";
#endif


#include "lsof.h"

_PROTOTYPE(static void read_mi,(struct stdata *s));
_PROTOTYPE(static int read_nfn,(caddr_t na, caddr_t fa, struct fifonode *f));
_PROTOTYPE(static int read_nhn,(caddr_t na, caddr_t ha, struct hsnode *h));
_PROTOTYPE(static int read_nin,(caddr_t na, caddr_t ia, struct inode *i));
_PROTOTYPE(static int read_npn,(caddr_t na, caddr_t pa, struct pcnode *p));
_PROTOTYPE(static int read_nrn,(caddr_t na, caddr_t ra, struct rnode *r));
_PROTOTYPE(static int read_nsn,(caddr_t na, caddr_t sa, struct snode *s));
_PROTOTYPE(static int read_ntn,(caddr_t na, caddr_t ta, struct tmpnode *t));
_PROTOTYPE(static int read_nvn,(caddr_t na, caddr_t va, struct vnode *v));
_PROTOTYPE(static int vop2ty,(unsigned long op));

#if	defined(HASPROCFS)
_PROTOTYPE(static int read_npi,(caddr_t na, struct vnode *v, struct pid *pids));
#endif	/* defined(HASPROCFS) */

#if	defined(solaris)
_PROTOTYPE(static char isvlocked,(struct vnode *va));
_PROTOTYPE(static int read_ncn,(caddr_t na, caddr_t ca, struct cnode *cn));
_PROTOTYPE(static int read_cni,(struct snode *s, struct vnode *rv,
	struct vnode *v, struct snode *rs, struct dev_info *di, char *din,
	int dinl));
#endif	/* defined(solaris) */


#if	defined(solaris)
/*
 * isvlocked() - is a vnode locked
 */

static char
isvlocked(va)
	struct vnode *va;		/* local vnode address */
{

	struct filock f;
	off_t l;

	if (va->v_filocks == NULL)
		return(' ');
	if (kread((KA_T)va->v_filocks, (char *)&f, sizeof(f)))
		return(' ');
	if (f.set.l_whence == 0 && f.set.l_start == 0 && f.set.l_len == 0)
		l = 1;
	else
		l = 0;
	switch (f.set.l_type & (F_RDLCK | F_WRLCK)) {

	case F_RDLCK:
		return((l) ? 'R' : 'r');
	case F_WRLCK:
		return((l) ? 'W' : 'w');
	case (F_RDLCK + F_WRLCK):
		return('*');
	}
	return(' ');
}
#endif	/* defined(solaris) */


/*
 * process_node() - process vnode
 */

void
process_node(va)
	caddr_t va;			/* vnode kernel space address */
{
	char dev_ch[32];
	dev_t dev;
	struct fifonode f;
	struct hsnode h;
	struct inode i;
	struct vfs kv;
	struct vfs *la = NULL;
	int ni = 0;
	struct pcnode pc;
	struct rnode r;
	struct vnode *realvp = NULL;
	struct vnode rv;
	struct snode s;
	struct tmpnode t;
	char *tn;
	char *ty;
	enum vtype type;
	struct vnode v;
	struct l_vfs *vfs;
	struct stdata *vs;
	int vty;

#if	defined(solaris)
	struct clone *c;
	struct cnode cn;
	struct dev_info di;
	char din[DINAMEL];
	struct filock fi;
	struct pcfs pcfs;
	struct snode rs;

# if	defined(HASPROCFS)
	struct procfsid *pfi;
	struct pid pids;
# endif	/* defined(HASPROCFS) */
#else	/* !defined(solaris) */
	long inum;
#endif	/* defined(solaris) */

/*
 * Read the vnode.
 */
	if ( ! va) {
		enter_nm("no vnode address");
		return;
	}
	if (readvnode((caddr_t)va, &v)) {
                enter_nm(Namech);
                return;
        }

#if	defined(HASNCACHE)
	Lf->na = (KA_T)va;
#endif	/* defined(HASNCACHE) */

	vs = v.v_stream;

#if	defined(solaris)
/*
 * Check for a Solaris socket.
 */
	if (vs) {
		for (c = Clone; c; c = c->next) {
		    if (c->n && major(v.v_rdev) == minor(c->rdev)) {
			process_socket(vs, &c->nm[c->n]);
			return;
		    }
		}
	}
/*
 * Obtain the Solaris virtual file system structure.
 */
	if (v.v_vfsp && kread((KA_T)v.v_vfsp, (char *)&kv, sizeof(kv))) {

vfs_read_error:

		(void) sprintf(Namech, "vnode at %#x: can't read vfs: %#x",
			va, v.v_vfsp);
		enter_nm(Namech);
		return;
	}
/*
 * Derive the virtual file system structure's device number from
 * its file system ID for NFS and High Sierra file systems.
 */
	if (kv.vfs_fstype > 1 && kv.vfs_fstype <= Fsinfomax) {
		if (strcmp(Fsinfo[kv.vfs_fstype - 1], "nfs") == 0
		||  strcmp(Fsinfo[kv.vfs_fstype - 1], "hsfs") == 0)
			kv.vfs_dev = (dev_t)kv.vfs_fsid.val[0];
	}
/*
 * Determine the Solaris vnode type.
 */
	if ((Ntype = vop2ty((unsigned long)v.v_op)) < 0)
		Ntype = N_REGLR;
	vty = Ntype;
	if (v.v_type == VFIFO)
		Ntype = N_FIFO;
	else if (vs) {
		Ntype = vty = N_STREAM;
		Lf->is_stream = 1;
	}
/*
 * See if this Solaris node is served by spec_vnodeops.
 */
	if (Nl[X_SPEC_OPS].n_value
	&&  Nl[X_SPEC_OPS].n_value == (unsigned long)v.v_op) 
		Ntype = N_SPEC;

/*
 * Determine the Solaris lock state.
 */
	Lf->lock = isvlocked(&v);
/*
 * Establish the Solaris local virtual file system structure.
 */
	if (v.v_vfsp == NULL)
		vfs = NULL;
	else if ((vfs = readvfs(v.v_vfsp, &kv)) == NULL)
		goto vfs_read_error;
#else	/* !defined(solaris) */

/*
 * Determine the non-Solaris vnode type.
 */
	if ((Ntype = vop2ty((unsigned long)v.v_op)) < 0) {
		if (v.v_type == VFIFO)
			Ntype = N_FIFO;
		else if (v.v_type == VCHR && !v.v_vfsp && vs) {
			Ntype = N_STREAM;
			Lf->is_stream = 1;
		} else
			Ntype = N_REGLR;
	}
/*
 * See if this non-Solaris node is served by _spec_vnodeops.
 */
	vty = Ntype;
	if (Nl[X_SPEC_OPS].n_value
	&&  (unsigned long)v.v_op == Nl[X_SPEC_OPS].n_value)
		Ntype = N_SPEC;
/*
 * Read the virtual file system structures for High Sierra and PC file
 * system vnodes to get major/minor device numbers from word 0 of the
 * file system ID array.
 */
	if (Ntype == N_HSFS || Ntype == N_PCFS) {
		if (!v.v_vfsp
		||  kread((KA_T)v.v_vfsp, (char *)&kv, sizeof(kv))) {
			(void) sprintf(Namech,
				"vnode at %#x: can't read vfs: %#x",
				va, v.v_vfsp);
			enter_nm(Namech);
			return;
		}
		la = &kv;
	}
/*
 * Determine the non-Solaris lock type.
 */
	if (v.v_shlockc || v.v_exlockc) {
		if (v.v_shlockc && v.v_exlockc)
			Lf->lock = '*';
		else if (v.v_shlockc)
			Lf->lock = 'R';
		else
			Lf->lock = 'W';
	}
/*
 * Establish the local virtual file system structure.
 */
	if (v.v_vfsp == NULL)
		vfs = NULL;
	else if ((vfs = readvfs(v.v_vfsp, la)) == NULL) {
                (void) sprintf(Namech, "vnode at %#x: can't read vfs: %#x",
			va, v.v_vfsp);
                enter_nm(Namech);
		return;
	}
#endif	/* defined(solaris) */

/*
 * Read the cnode, fifonode, inode, pcnode, rnode, snode or tmpnode.
 */
	switch (Ntype) {
	case N_SPEC:
	
	/*
	 * A N_SPEC node is node that resides in in an underlying file system
	 * type -- e.g. NFS, HSFS.  Its vnode points to an snode.  Subsequent
	 * node structures are implied by the underlying node type.
	 */
		if (read_nsn(va, v.v_data, &s))
			return;
		realvp = s.s_realvp;

#if	defined(solaris)
		if (!realvp && s.s_commonvp) {
		    if (read_cni(&s, &rv, &v, &rs, &di, din, sizeof(din))
		    == 1)
			return;
		    if (!rv.v_stream) {
			if (din[0]) {
			    (void) strcpy(Namech, din);
			    Lf->is_com = 1;
			}
			ni = 1;
			break;
		    }
		}
#endif	/* defined(solaris) */

		if (!realvp) {
		    ni = 1;

		/*
		 * If the snode lacks a real vnode (and also lacks a
		 * Solaris common vnode), it's original type is N_STREAM,
		 * and it has a stream pointer, get the module names.
		 */
		    if (vty == N_STREAM && vs) {
			(void) strcpy(Lf->iproto, "STR");
			read_mi(vs);
			vs = NULL;
		    }
		}
		break;

#if	defined(solaris)
	case N_CACHE:
		if (read_ncn(va, v.v_data, &cn))
			return;
		break;
#endif	/* defined(solaris) */

	case N_NFS:
		if (read_nrn(va, v.v_data, &r))
			return;
		break;
	case N_FIFO:
		if (read_nfn(va, v.v_data, &f))
		    return;

#if	defined(solaris)
	/*
	 * Solaris FIFO vnodes are linked to a fifonode.
	 * Non-pipe fifonodes are linked to a vnode thorough fn_realvp.
	 */
		if (!(realvp = f.fn_realvp)) {
			Lf->inode = (unsigned long)f.fn_ino;
			Lf->inp_ty = 1;
			(void) sprintf(dev_ch, "0x%08x", v.v_data);
			enter_dev_ch(dev_ch);
			if (f.fn_flag & ISPIPE)
				(void) strcpy(Namech, "PIPE");
			if (f.fn_mate)
				(void) sprintf(endnm(), "->0x%08x", f.fn_mate);
			break;
		}
#else	/* !defined(solaris) */
	/*
	 * A non-Solaris FIFO vnode is linked to a fifonode.  The s_realvp
	 * pointer of the snode contained in the fifonode points to a vnode
	 * that points to the inode.  (Whew!)
	 */
		if ((realvp = f.fn_snode.s_realvp) == NULL) {
		    (void) sprintf(Namech,
			"fifonode at %#x has no successor", v.v_data);
		    enter_nm(Namech);
		    return;
		}
#endif	/* defined(solaris) */
		
		break;

	case N_HSFS:
		if (read_nhn(va, v.v_data, &h))
		    return;
		break;
	case N_PCFS:
		if (read_npn(va, v.v_data, &pc))
		    return;
		break;

#if	defined(HASPROCFS)
	case N_PROC:
		ni = 1;
		if (read_npi(va, &v, &pids))
		    return;
		break;
#endif	/* defined(HASPROCFS) */

	case N_TMP:
		if (read_ntn(va, v.v_data, &t))
		    return;
		break;
	case N_STREAM:
		if (read_nsn(va, v.v_data, &s))
			return;
		if (vs) {
			(void) strcpy(Lf->iproto, "STR");
			read_mi(vs);
			vs = NULL;
		}
		break;
	case N_REGLR:
	default:
		if (read_nin(va, v.v_data, &i))
		    return;
	}
/*
 * If the node has a real vnode pointer, follow it.
 */
	if (realvp) {
		if (read_nvn(v.v_data, (caddr_t)realvp, &v))
			return;
	/*
	 * If the original vnode type is N_STREAM, and if there is
	 * a stream pointer, get the module names.
	 */
		if (vty == N_STREAM && vs) {
			(void) strcpy(Lf->iproto, "STR");
			read_mi(vs);
			vs = NULL;
		}
	/*
	 * Get the real vnode's type.
	 */
		if ((vty = vop2ty((unsigned long)v.v_op)) < 0) {

#if	defined(solaris)
			if (Ntype != N_FIFO && vs)
#else	/* !defined(solaris) */
			if (v.v_type == VCHR && !v.v_vfsp && v.v_stream)
#endif	/* defined(solaris) */

				vty = N_STREAM;
			else
				vty = N_REGLR;
		}
	/*
	 * Base further processing on the "real" vnode.
	 */
		switch (vty) {
	
#if	defined(solaris)
		case N_CACHE:
			if (read_ncn(va, v.v_data, &cn))
				return;
			break;
#endif	/* defined(solaris) */
	
		case N_HSFS:
			if (read_nhn(va, v.v_data, &h))
				return;
			break;
		case N_NFS:
			if (read_nrn(va, v.v_data, &r))
				return;
			break;
		case N_PCFS:
			if (read_npn(va, v.v_data, &pc))
				return;
			break;
		case N_STREAM:
			if (vs) {
				(void) strcpy(Lf->iproto, "STR");
				read_mi(vs);
				vs = NULL;
			}
			break;
		case N_TMP:
			if (read_ntn(va, v.v_data, &t))
				return;
			break;
		case N_REGLR:
		default:
			if (read_nin(va, v.v_data, &i))
				return;
		}
	}
/*
 * Get device and type for printing.
 */
	switch (Ntype == N_FIFO ? vty : Ntype) {

#if	defined(solaris)
	case N_CACHE:
	case N_HSFS:
	case N_PCFS:
		dev = kv.vfs_dev;
#else	/* !defined(solaris) */
	case N_HSFS:
	case N_PCFS:
		dev = kv.vfs_fsid.val[0] & 0xffff;
#endif	/* defined(solaris) */

		break;
	case N_NFS:
		dev = r.r_attr.va_fsid;
		break;

#if	defined(HASPROCFS)
	case N_PROC:
		dev = kv.vfs_dev;
		break;
#endif	/* defined(HASPROCFS) */

	case N_SPEC:
		dev = s.s_dev;
		Ntype = vty;
		break;
	case N_STREAM:
		dev = s.s_dev;
		break;
	case N_TMP:
		dev = t.tn_attr.va_fsid;
		break;
	default:
		dev = (v.v_type == VCHR || v.v_type == VBLK || ni) ? v.v_rdev
								   : i.i_dev;
	}
	type = v.v_type;
	if (vfs && vfs->dir == NULL)
		(void) completevfs(vfs, &dev);
/*
 * Obtain the inode number.
 */
	switch (vty) {

#if	defined(solaris)
	case N_CACHE:
		Lf->inode = (unsigned long)cn.c_fileno;
		Lf->inp_ty = 1;
		break;
#endif	/* defined(solaris) */

	case N_HSFS:

#if	defined(solaris)
		Lf->inode = (unsigned long)h.hs_nodeid;
#else	/* !defined(solaris) */
		Lf->inode = (unsigned long)h.hs_dirent.ext_lbn;
#endif	/* defined(solaris) */

		Lf->inp_ty = 1;
		break;

	case N_NFS:
		Lf->inode = (unsigned long)r.r_attr.va_nodeid;
		Lf->inp_ty = 1;
		break;

#if	defined(HASPROCFS)
	case N_PROC:

	/*
	 * The proc file system inode number is defined when the
	 * prnode is read.
	 */
		break;
#endif	/* defined(HASPROCFS) */

	case N_PCFS:


#if	defined(solaris)
		if (kv.vfs_data
		&& kread((KA_T)kv.vfs_data, (char *)&pcfs, sizeof(pcfs)) == 0)
		{
			Lf->inode = (unsigned long)pc_makenodeid(pc.pc_eblkno,
						   pc.pc_eoffset,
						   &pc.pc_entry,
						   pcfs.pcfs_entps);
			Lf->inp_ty = 1;
		}
#else	/* !defined(solaris) */
		Lf->inode = (unsigned long)pc_makenodeid(pc.pc_eblkno,
							 pc.pc_eoffset,
							 &pc.pc_entry);
		Lf->inp_ty = 1;
#endif	/* defined(solaris) */

		break;

	case N_REGLR:
		if (!ni) {
			Lf->inode = (unsigned long)i.i_number;

#if	!defined(solaris)
			if (type != VBLK || Lf->inode)
#endif	/* !defined(solaris) */

			Lf->inp_ty = 1;
		}
		break;
	case N_TMP:
		Lf->inode = (unsigned long)t.tn_attr.va_nodeid;
		Lf->inp_ty = 1;
		break;
	}
/*
 * Obtain the file size.
 */
	if (Foffset)
		Lf->off_def = 1;
	else {
		switch (Ntype) {

#if	defined(solaris)
		case N_CACHE:
			Lf->sz = cn.c_size;
			Lf->sz_def = 1;
			break;
#endif	/* defined(solaris) */

		case N_HSFS:
			Lf->sz = h.hs_dirent.ext_size;
			Lf->sz_def = 1;
			break;
		case N_FIFO:
			if (!Fsize)
				Lf->off_def = 1;
			break;
		case N_NFS:

			if ((type == VCHR || type == VBLK) && !Fsize)
				Lf->off_def = 1;
			else {

#if	defined(solaris)
				Lf->sz = r.r_size;
#else	/* !defined(solaris) */
				Lf->sz = r.r_attr.va_size;
#endif	/* defined(solaris) */

				Lf->sz_def = 1;
			}
			break;

		case N_PCFS:
			Lf->sz = pc.pc_size;
			Lf->sz_def = 1;
			break;

#if	defined(HASPROCFS)
		case N_PROC:

		/*
		 * The proc file system size is defined when the
		 * prnode is read.
		 */
			break;
#endif	/* defined(HASPROCFS) */

		case N_REGLR:
			if (type == VREG || type == VDIR) {
				if (!ni) {
					Lf->sz = i.i_size;
					Lf->sz_def = 1;
				}
			} else if ((type == VCHR || type == VBLK) && !Fsize)
				Lf->off_def = 1;
			break;
		case N_STREAM:
			if (!Fsize)
				Lf->off_def = 1;
			break;
		case N_TMP:
			Lf->sz = t.tn_attr.va_size;
			Lf->sz_def = 1;
		}
	}
/*
 * Record an NFS selection.
 */
	if (Ntype == N_NFS && Fnfs)
		Lf->sf |= SELNFS;
/*
 * Save the file system names.
 */
	if (vfs) {
		Lf->fsdir = vfs->dir;
		Lf->fsdev = vfs->fsname;
	}

#if	!defined(solaris)
/*
 * If this is the current working or root directory, and there is a path
 * name, use it.
 */
	if (*Cwd && strcmp(Lf->fd, CWD) == 0)
		(void) strcpy(Namech, Cwd);
	else if (*Rtd && strcmp(Lf->fd, RTD) == 0)
		(void) strcpy(Namech, Rtd);
#endif	/* !defined(solaris) */

/*
 * Format the vnode type, and possibly the device name.
 */
	switch (type) {

	case VNON:
		ty ="VNON";
		break;
	case VREG:
	case VDIR:
		ty = (type == VREG) ? "VREG" : "VDIR";
		Lf->dev_def = 1;
		Lf->dev = dev;
		break;
	case VBLK:
		ty = "VBLK";
		Lf->dev_def = 1;
		Lf->dev = dev;
		break;
	case VCHR:
		Lf->dev = dev;
		Lf->dev_def = 1;
		ty = "VCHR";
		if (Lf->is_stream == 0 && Lf->is_com == 0)
			Lf->is_chr_dev = 1;
		break;
	case VLNK:
		ty = "VLNK";
		break;

#if	defined(VSOCK)
	case VSOCK:
		ty = "SOCK";
		break;
#endif	/* defined(VSOCK) */

	case VBAD:
		ty = "VBAD";
		break;
	case VFIFO:
		if (!Lf->dev_ch || Lf->dev_ch[0] == '\0') {
			Lf->dev = dev;
			Lf->dev_def = 1;
		}
		ty = "FIFO";
		break;
	default:
		if (type > 9999)
			(void) sprintf(Lf->type, "*%03d", type % 1000);
		else
			(void) sprintf(Lf->type, "%4d", type);
		(void) strcpy(Namech, "unknown type");
		ty = NULL;
	}
	if (ty)
		(void) strcpy(Lf->type, ty);

#if	defined(solaris)
/*
 * If this a Solaris common vnode/snode void some information.
 */
	if (Lf->is_com)
		Lf->sz_def = Lf->inp_ty = 0;
#endif	/* defined(solaris) */

/*
 * If this is a VCHR file and it's missing an inode number, try to
 * supply one.
 */

#if	defined(solaris)
	if (Lf->inp_ty == 0 && type == VCHR && Lf->dev_def)
#else	/* !defined(solaris) */
	if (Lf->inp_ty == 0 && (type == VCHR || type == VBLK) && Lf->dev_def)
#endif	/* defined(solaris) */

		find_ch_ino();
	if (Lf->inp_ty == 0 && Lf->is_stream && strcmp(Lf->iproto, "STR") == 0)
		Lf->inp_ty = 2;
/*
 * Test for specified file.
 */

#if	defined(HASPROCFS)
	if (Ntype == N_PROC) {
		if (Procsrch)
			Lf->sf |= SELNM;
		else {
			for (pfi = Procfsid; pfi; pfi = pfi->next) {
				if (pfi->pid == pids.pid_id) {
					Lf->sf |= SELNM;
					break;
				}
			}
		}
	} else
#endif	/* defined(HASPROCFS) */

	{
		if (Sfile && is_file_named(NULL, Ntype, type))
			Lf->sf |= SELNM;
	}
/*
 * Enter name characters.
 */
	if (Namech[0])
		enter_nm(Namech);
}


#if	defined(solaris)
/*
 * read_cni() - read common snode information
 */

static int
read_cni(s, rv, v, rs, di, din, dinl)
	struct snode *s;		/* starting snode */
	struct vnode *rv;		/* "real" vnode receiver */
	struct vnode *v;		/* starting vnode */
	struct snode *rs;		/* "real" snode receiver */
	struct dev_info *di;		/* dev_info structure receiver */
	char *din;			/* device info name receiver */
	int dinl;			/* sizeof(*din) */
{
	if (read_nvn(v->v_data, (caddr_t)s->s_commonvp, rv))
		return(1);
	if (read_nsn((caddr_t)s->s_commonvp, rv->v_data, rs))
		return(1);
	*din = '\0';
	if (rs->s_dip) {
	    if (kread((KA_T)rs->s_dip, (char *)di, sizeof(struct dev_info))) {
		(void) sprintf(Namech,
		    "common snode at %#x: no dev info: %#x",
		    rv->v_data, rs->s_dip);
		enter_nm(Namech);
		return(1);
	    }
	    if (di->devi_name
	    &&  kread((KA_T)di->devi_name, din, dinl-1) == 0)
		din[dinl-1] = '\0';
	}
	return(0);
}
#endif	/* defined(solaris) */


/*
 * read_mi() - read stream's module information
 */

static void
read_mi(s)
	struct stdata *s;			/* stream pointer */
{
	int i, j, k, nl;
	KA_T ka, qp;
	struct module_info mi;
	char mn[STRNML];
	struct stdata sd;
	struct queue q;
	struct qinit qi;
/*
 * If there is no stream pointer, or we can't read the stream head,
 * return.
 */
	if (!s)
		return;
	if (kread((KA_T)s, (char *)&sd, sizeof(sd))) {
		(void) sprintf(Namech, "can't read stream head: %#x", s);
		return;
	}
/*
 * Follow the stream head to each of its queue structures, retrieving the
 * module names from each queue's q_info->qi_minfo->mi_idname chain of
 * structures.  Separate each additional name from the previous one with
 * "->".
 *
 * Ignore failures to read all but queue structure chain entries.
 *
 * Ignore module names that end in "head".
 */
	k = 0;
	Namech[0] = '\0';
	nl = sizeof(mn) - 1;
	mn[nl] = '\0';
	qp = (KA_T)sd.sd_wrq;
	for (i = 0; qp && i < 20; i++, qp = (KA_T)q.q_next) {
		if (kread(qp, (char *)&q, sizeof(q)))
			break;
		if ((ka = (KA_T)q.q_qinfo) == (KA_T)NULL
		||  kread(ka, (char *)&qi, sizeof(qi)))
			continue;
		if ((ka = (KA_T)qi.qi_minfo) == (KA_T)NULL
		||  kread(ka, (char *)&mi, sizeof(mi)))
			continue;
		if ((ka = (KA_T)mi.mi_idname) == (KA_T)NULL
		||  kread(ka, mn, nl))
			continue;
		if ((j = strlen(mn)) < 1)
			continue;
		if (j >= 4 && strcmp(&mn[j - 4], "head") == 0)
			continue;
		if (k) {
			if ((k + 2) > (MAXPATHLEN - 1))
				break;
			(void) strcpy(&Namech[k], "->");
			k += 2;
		}
		if ((k + j) > (MAXPATHLEN - 1))
			break;
		(void) strcpy(&Namech[k], mn);
		k += j;
	}
}


#if	defined(solaris)
/*
 * read_ncn(na, ca, cn) - read node's cache node
 */

static int
read_ncn(na, ca, cn)
	caddr_t na;			/* containing node's address */
	caddr_t ca;			/* cache node address */
	struct cnode *cn;		/* cache node receiver */
{
	if (!ca || kread((KA_T)ca, (char *)cn, sizeof(struct cnode))) {
		(void) sprintf(Namech,
			"node at %#x: can't read cnode: %#x", na, cn);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}
#endif	/* defined(solaris) */


/*
 * read_nfn() - read node's fifonode
 */

static int
read_nfn(na, fa, f)
	caddr_t na;			/* containing node's address */
	caddr_t fa;			/* fifonode address */
	struct fifonode *f;		/* fifonode receiver */
{
	if (!fa || readfifonode(fa, f)) {
		(void) sprintf(Namech,
			"node at %#x: can't read fifonode: %#x", na, fa);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}


/*
 * read_nhn() - read node's High Sierra node
 */

static int
read_nhn(na, ha, h)
	caddr_t na;			/* containing node's address */
	caddr_t ha;			/* hsnode address */
	struct hsnode *h;		/* hsnode receiver */
{
	if (!ha || readhsnode(ha, h)) {
		(void) sprintf(Namech,
			"node at %#x: can't read hsnode: %#x", na, ha);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}


/*
 * read_nin() - read node's inode
 */

static int
read_nin(na, ia, i)
	caddr_t na;			/* containing node's address */
	caddr_t ia;			/* kernel inode address */
	struct inode *i;		/* inode receiver */
{
	if (!ia || readinode((struct inode *)ia, i)) {
		(void) sprintf(Namech,
			"node at %#x: can't read inode: %#x", na, ia);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}


#if	defined(HASPROCFS)
/*
 * read_npi() - read node's /proc file system information
 */

static int
read_npi(na, v, pids)
	caddr_t na;			/* containing node's address */
	struct vnode *v;		/* containing vnode */
	struct pid *pids;		/* pid structure receiver */
{
	struct as as;
	struct proc p;
	struct prnode pr;

	if (!v->v_data || kread((KA_T)v->v_data, (char *)&pr, sizeof(pr))) {
		(void) sprintf(Namech,
			"node at %#x: can't read prnode: %#x", na, v->v_data);
		enter_nm(Namech);
		return;
	}
	if (pr.pr_proc == NULL) {
		if (v->v_type == VDIR) {
			(void) sprintf(Namech, "/%s", HASPROCFS);
			enter_nm(Namech);
			Lf->inode = PR_ROOTINO;
			Lf->inp_ty = 1;
		} else {
			(void) sprintf(Namech, "/%s/???", HASPROCFS);
			enter_nm(Namech);
			Lf->inp_ty = 0;
		}
		return(0);
	}
	if (kread((KA_T)pr.pr_proc, (char *)&p, sizeof(p))) {
		(void) sprintf(Namech,
			"prnode at %#x: can't read proc: %#x",
			v->v_data, pr.pr_proc);
		enter_nm(Namech);
		return(1);
	}
	if (!p.p_pidp
	||  kread((KA_T)p.p_pidp, (char *)pids, sizeof(struct pid))) {
		(void) sprintf(Namech,
			"proc struct at %#x: can't read pid: %#x",
			pr.pr_proc, p.p_pidp);
		enter_nm(Namech);
		return(1);
	}
	if (p.p_as && kread((KA_T)p.p_as, (char *)&as, sizeof(as)) == 0) {
		Lf->sz = as.a_size;
		Lf->sz_def = 1;
	}
	(void) sprintf(Namech, "/%s/%0*d", HASPROCFS, PNSIZ, pids->pid_id);
	Lf->inode = (unsigned long)ptoi(pids->pid_id);
	Lf->inp_ty = 1;
	enter_nm(Namech);
	return(0);
}
#endif	/* defined(HASPROCFS) */


/*
 * read_npn() - read node's pcnode
 */

static int
read_npn(na, pa, p)
	caddr_t na;			/* containing node's address */
	caddr_t pa;			/* pcnode address */
	struct pcnode *p;		/* pcnode receiver */
{
	if (pa == NULL || kread((KA_T)pa, (char *)p, sizeof(struct pcnode))) {
		(void) sprintf(Namech,
			"node at %#x: can't read pcnode: %#x", na, pa);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}


/*
 * read_nrn() - read node's rnode
 */

static int
read_nrn(na, ra, r)
	caddr_t na;			/* containing node's address */
	caddr_t ra;			/* rnode address */
	struct rnode *r;		/* rnode receiver */
{
	if (!ra || readrnode(ra, r)) {
		(void) sprintf(Namech,
			"node at %#x: can't read rnode: %#x", na, ra);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}


/*
 * read_nsn() - read node's snode
 */

static int
read_nsn(na, sa, s)
	caddr_t na;			/* containing node's address */
	caddr_t sa;			/* snode address */
	struct snode *s;		/* snode receiver */
{
	if (!sa || readsnode((caddr_t)sa, s)) {
		(void) sprintf(Namech,
			"node at %#x: can't read snode: %#x", na, sa);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}


/*
 * read_ntn() - read node's tmpnode
 */

static int
read_ntn(na, ta, t)
	caddr_t na;			/* containing node's address */
	caddr_t ta;			/* tmpnode address */
	struct tmpnode *t;		/* tmpnode receiver */
{
	if (!ta || readtnode(ta, t)) {
		(void) sprintf(Namech,
			"node at %#x: can't read tnode: %#x", na, ta);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}


/*
 * read_nvn() - read node's vnode
 */

static int
read_nvn(na, va, v)
	caddr_t na;			/* node's address */
	caddr_t va;			/* vnode address */
	struct vnode *v;		/* vnode receiver */
{
	if (readvnode(va, v)) {
		(void) sprintf(Namech,
			"node at %#x: can't read real vnode: %#x", na, va);
		enter_nm(Namech);
		return(1);
	}
	return(0);
}


/*
 * vop2ty() - convert vnode operation switch address to internal type
 */

static int
vop2ty(op)
	unsigned long op;		/* vnode's v_op element */
{
	if (Nl[X_NFS_OPS].n_value && op == Nl[X_NFS_OPS].n_value)
		return(N_NFS);
	else if (Nl[X_TMP_OPS].n_value && op == Nl[X_TMP_OPS].n_value)
		return(N_TMP);
	else if (Nl[X_HSFS_OPS].n_value && op == Nl[X_HSFS_OPS].n_value)
		return(N_HSFS);
	else if ((Nl[X_PC_DOPS].n_value && op == Nl[X_PC_DOPS].n_value)
	     ||  (Nl[X_PC_FOPS].n_value && op == Nl[X_PC_FOPS].n_value))
		return(N_PCFS);

#if	defined(X_CACHE_OPS)
	else if (Nl[X_CACHE_OPS].n_value && op == Nl[X_CACHE_OPS].n_value)
		return(N_CACHE);
#endif	/* defined(X_CACHE_OPS) */

#if	defined(X_FIFO_OPS)
	else if (Nl[X_FIFO_OPS].n_value && op == Nl[X_FIFO_OPS].n_value)
		return(N_FIFO);
#endif	/* defined(X_FIFO_OPS) */


# if	defined(X_PROC_OPS)
	else if (Nl[X_PROC_OPS].n_value && op == Nl[X_PROC_OPS].n_value)
		return(N_PROC);
#endif	/* defined(X_PROC_OPS) */

	return(-1);
}
