/*
 * store.c - common global storage for lsof
 */


/*
 * Copyright 1994 Purdue Research Foundation, West Lafayette, Indiana
 * 47907.  All rights reserved.
 *
 * Written by Victor A. Abell
 *
 * This software is not subject to any license of the American Telephone
 * and Telegraph Company or the Regents of the University of California.
 *
 * Permission is granted to anyone to use this software for any purpose on
 * any computer system, and to alter it and redistribute it freely, subject
 * to the following restrictions:
 *
 * 1. Neither the authors nor Purdue University are responsible for any
 *    consequences of the use of this software.
 *
 * 2. The origin of this software must not be misrepresented, either by
 *    explicit claim or by omission.  Credit to the authors and Purdue
 *    University must appear in documentation and sources.
 *
 * 3. Altered versions must be plainly marked as such, and must not be
 *    misrepresented as being the original software.
 *
 * 4. This notice may not be removed or altered.
 */

#ifndef lint
static char copyright[] =
"@(#) Copyright 1994 Purdue Research Foundation.\nAll rights reserved.\n";
static char *rcsid = "$Id: store.c,v 1.9 95/02/13 14:32:01 abe Exp $";
#endif


#include "lsof.h"


/*
 * Global storage definitions
 */

char *Core = NULL;		/* core file path */

#if	defined(HASDCACHE)
unsigned DCcksum;		/* device cache file checksum */
int DCfd = -1;			/* device cache file descriptor */
FILE *DCfs = NULL;		/* stream pointer for DCfd */
int DCstate = 3;		/* device cache state:
				 *	0 = ignore (-Di)
				 *	1 = build (-Db[path])
				 *	2 = read; don't rebuild (-Dr[path])
				 *	3 = update; read and rebuild if
				 *	    necessary (-Du[path])
				 */
char *DCpath = HASDCACHE;	/* device cache path */
#endif	/* defined(HASDCACHE) */

struct dev *Devtp = NULL;	/* device table pointer */
int Fand = 0;			/* -a flag status */
int Fblock = 0;			/* -b flag status */

#if	defined(HASNCACHE)
int Fncache = 1;		/* -C flag status */
#endif	/* defined(HASNCACHE) */

int Fdevmsg = 0;		/* -d flag status */
int Fhelp = 0;			/* -h flag status */
int Fhost = 1;			/* -H flag status */
int Fmount = 0;			/* -m flag status */
int Fnet = 0;			/* -n flag status */
int Fnfs = 0;			/* -N flag status */
int Foffset = 0;		/* -o flag status */
int Forder = 1;			/* -O flag status */
int Fport = 1;			/* -P flag status */
int Fpgrp = 0;			/* -g flag status */
int Fsize = 0;			/* -s flag status */
int Fterse = 0;			/* -t flag status */
int Funix = 0;			/* -U flag status */
int Futol = 1;			/* -l flag status */
int Fwarn = 0;			/* -w flag status */
int Hdr = 0;			/* header print status */
struct lfile *Lf = NULL;	/* current local file structure */
struct lproc *Lp = NULL;	/* current local process table entry */
struct lproc *Lproc = NULL;	/* local process table */

struct mounts *Mtab = NULL;	/* mounted devices */

#if	defined(HASPROCFS)
struct mounts *Mtprocfs = NULL;	/* /proc mount entry */
#endif

int Mxpgrp = 0;			/* maximum process groip ID table entries */
int Mxpid = 0;			/* maximum PID table entries */
int Mxuid = 0;			/* maximum UID table entries */
int Mypid;			/* lsof's process ID */

#if	defined(HASSECURITY)
uid_t Myuid;			/* real UID of this lsof process */
#endif

char *Namech = NULL;		/* name characters for printing */
int Ndev = 0;			/* number of device table entries */
int Nlproc = 0;			/* number of entries in Lproc[] */
char *Nmlst = NULL;		/* namelist file path */
int Npgrp = 0;			/* -g flag count */
int Npid = 0;			/* -p flag count */
int Npuns;			/* number of unselected PIDs (starts at Npid) */
int Ntype;			/* node type (see N_* symbols) */
int Nuid = 0;			/* -u flag count */
struct nwad *Nwad = NULL;	/* list of network addresses */
struct lfile *Plf = NULL;	/* previous local file structure */
char *Pn;			/* program name */

#if	defined(HASPROCFS)
struct procfsid *Procfsid = NULL;
				/* proc file system PID search table */
int Procsrch = 0;		/* 1 if searching for any proc file system
				 * file */
#endif

struct dev **Sdev = NULL;	/* pointer to device table pointers, sorted
				 * by device */
int Selall = 1;			/* all processes are selected (default) */
int Selflags = 0;		/* selection flags -- see SEL* in lsof.h */
struct sfile *Sfile = NULL;	/* chain of files to search for */
int *Spgrp = NULL;		/* process group IDs to search for */
int *Spid = NULL;		/* Process IDs to search for */
uid_t *Suid = NULL;		/* User IDs to search for */
int TmLimit = TMLIMIT;		/* Readlink() and stat() timeout (seconds) */
